package be.iminds.ilabt.jfed.espec.filefetcher;

import be.iminds.ilabt.jfed.espec.bundle.BundleFetcher;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.bundle.LocalArchiveESpecBundle;
import be.iminds.ilabt.jfed.espec.model.ExperimentSpecification;
import be.iminds.ilabt.jfed.espec.model.UploadSpec;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import org.testng.annotations.Test;

import javax.annotation.Nonnull;

import java.io.File;
import java.net.URL;
import java.nio.file.Path;
import java.util.Iterator;
import java.util.List;

import static org.testng.Assert.*;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;

import static org.mockito.Mockito.mock;

public class ExperimentSpecificationFileManagerTest {
    public void testArchiveBundleFilenames(ESpecBundle bundle) throws Exception {
        List<String> filenames = bundle.getFileNames();
        System.out.println("Archive filenames="+filenames);
        assertThat(filenames, hasItem(equalTo("a")));
        assertThat(filenames, hasItem(equalTo("b")));
        assertThat(filenames, not(hasItem(equalTo("d/c"))));
        assertThat(filenames, not(hasItem(equalTo("d/e"))));
        assertThat(filenames, not(hasItem(equalTo("d/dd/f"))));
        assertThat(filenames, hasItem(anyOf(equalTo("d/"), equalTo("d"))));
        assertThat(filenames, not(hasItem(anyOf(equalTo("d/dd/"), equalTo("d/dd")))));
    }
    public void testArchiveBundleDirs(ESpecBundle bundle) throws Exception {
        assertTrue(!bundle.isDir("a"));
        assertTrue(!bundle.isDir("b"));
        assertTrue(!bundle.isDir("d/c"));
        assertTrue(!bundle.isDir("d/e"));
        assertTrue(!bundle.isDir("d/dd/f"));
        assertTrue(bundle.isDir("d/"));
        assertTrue(bundle.isDir("d/dd/"));
        assertTrue(bundle.isDir("d"));
        assertTrue(bundle.isDir("d/dd"));
    }
    public void testArchiveBundleSubDirs(ESpecBundle bundle) throws Exception {
        List<String> dirFilenamesD = bundle.getDirFiles("d");
        System.out.println("dirFilenamesD="+dirFilenamesD);
        assertThat(dirFilenamesD, hasItem(anyOf(equalTo("d/dd"), equalTo("d/dd/"))));
        assertThat(dirFilenamesD, hasItem(equalTo("d/c")));
        assertThat(dirFilenamesD, hasItem(equalTo("d/e")));
        assertThat(dirFilenamesD, not(hasItem(anyOf(equalTo("d"), equalTo("d/")))));
        assertThat(dirFilenamesD, not(hasItem(equalTo("a"))));
        assertThat(dirFilenamesD, not(hasItem(equalTo("b"))));
        assertThat(dirFilenamesD, not(hasItem(equalTo("d/dd/f"))));

        dirFilenamesD = bundle.getDirFiles("d/");
        assertThat(dirFilenamesD, hasItem(anyOf(equalTo("d/dd"), equalTo("d/dd/"))));
        assertThat(dirFilenamesD, hasItem(equalTo("d/c")));
        assertThat(dirFilenamesD, hasItem(equalTo("d/e")));
        assertThat(dirFilenamesD, not(hasItem(anyOf(equalTo("d/"), equalTo("d")))));
        assertThat(dirFilenamesD, not(hasItem(equalTo("a"))));
        assertThat(dirFilenamesD, not(hasItem(equalTo("b"))));
        assertThat(dirFilenamesD, not(hasItem(equalTo("d/dd/f"))));

        List<String> dirFilenamesDD = bundle.getDirFiles("d/dd");
        System.out.println("dirFilenamesDD="+dirFilenamesDD);
        assertThat(dirFilenamesDD, hasItem(equalTo("d/dd/f")));
        assertThat(dirFilenamesDD, not(hasItem(equalTo("d/c"))));
        assertThat(dirFilenamesDD, not(hasItem(equalTo("d/e"))));
        assertThat(dirFilenamesDD, not(hasItem(anyOf(equalTo("d/"), equalTo("d")))));
        assertThat(dirFilenamesDD, not(hasItem(anyOf(equalTo("d/dd"), equalTo("d/dd/")))));
        assertThat(dirFilenamesDD, not(hasItem(equalTo("a"))));
        assertThat(dirFilenamesDD, not(hasItem(equalTo("b"))));

        dirFilenamesDD = bundle.getDirFiles("d/dd/");
        assertThat(dirFilenamesDD, hasItem(equalTo("d/dd/f")));
        assertThat(dirFilenamesDD, not(hasItem(equalTo("d/c"))));
        assertThat(dirFilenamesDD, not(hasItem(equalTo("d/e"))));
        assertThat(dirFilenamesDD, not(hasItem(anyOf(equalTo("d/"), equalTo("d")))));
        assertThat(dirFilenamesDD, not(hasItem(anyOf(equalTo("d/dd"), equalTo("d/dd/")))));
        assertThat(dirFilenamesDD, not(hasItem(equalTo("a"))));
        assertThat(dirFilenamesDD, not(hasItem(equalTo("b"))));
    }

    public void testDirIterator(ESpecBundle bundle) throws Exception {
        ExperimentSpecification eSpec = bundle.getExperimentSpecification();
        ESpecLogger eSpecLogger = new ESpecLogger();
        ExperimentSpecificationFileManager man =
                new ExperimentSpecificationFileManager(eSpec, bundle, eSpecLogger, null, null);

        man.fetchAll();
        assertTrue(man.isReady());

        boolean sawD = false;
        boolean sawC = false;
        boolean sawE = false;
        boolean sawF = false;
        boolean sawDD = false;
        for (UploadSpec upload : eSpec.getUploads()) {
            if (man.isDir(upload.getSource()) && assureEndWithSlash(upload.getSource().getValue()).equals("d/")) {
                Iterator<ExperimentSpecificationFileManager.DirFileBytes> it = man.getDirBytes(upload.getSource());
                assert it.hasNext();
                while(it.hasNext()) {
                    ExperimentSpecificationFileManager.DirFileBytes dirFileBytes = it.next();
                    System.out.println("Iterate dirFileBytes -> "+dirFileBytes.getFullRelativePath()+"  -> base = "+dirFileBytes.getBaseFilename());
                    switch (dirFileBytes.getBaseFilename()) {
                        case "" : {
                            assertFalse(sawD);
                            assertTrue(dirFileBytes.isDir());
                            assertThat(dirFileBytes.getFullRelativePath(), is(equalTo("")));
                            sawD = true;
                            break;
                        }
                        case "c" : {
                            assertFalse(sawC);
                            assertTrue(sawD);
                            assertFalse(dirFileBytes.isDir());
                            assertThat(dirFileBytes.getContent(), is(equalTo(new byte[] { 'c', '\n' })));
                            assertThat(dirFileBytes.getFullRelativePath(), is(equalTo("c")));
                            sawC = true;
                            break;
                        }
                        case "e" : {
                            assertFalse(sawE);
                            assertTrue(sawD);
                            assertFalse(dirFileBytes.isDir());
                            assertThat(dirFileBytes.getContent(), is(equalTo(new byte[] { 'e', 'e', '\n' })));
                            assertThat(dirFileBytes.getFullRelativePath(), is(equalTo("e")));
                            sawE = true;
                            break;
                        }
                        case "f" : {
                            assertFalse(sawF);
                            assertTrue(sawDD);
                            assertFalse(dirFileBytes.isDir());
                            assertThat(dirFileBytes.getContent(), is(equalTo(new byte[] { 'f', 'f', 'f', '\n' })));
                            assertThat(dirFileBytes.getFullRelativePath(), is(equalTo("dd/f")));
                            sawF = true;
                            break;
                        }
                        case "dd/" : //fall through
                        case "dd" : {
                            assertFalse(sawF);
                            assertFalse(sawDD);
                            assertTrue(dirFileBytes.isDir());
                            assertThat(dirFileBytes.getFullRelativePath(), is(anyOf(equalTo("dd"), equalTo("dd/"))));
                            sawDD = true;
                            break;
                        }
                        default: {
                            assertTrue(false, "Bad base filename: "+dirFileBytes.getBaseFilename());
                            throw new RuntimeException("assert failed");
                        }
                    }
                }

            }
        }
        assertTrue(sawDD);
        assertTrue(sawC);
        assertTrue(sawE);
        assertTrue(sawF);
    }

    public void testSubDirIterator(ESpecBundle bundle) throws Exception {
        ExperimentSpecification eSpec = bundle.getExperimentSpecification();
        ESpecLogger eSpecLogger = new ESpecLogger();
        ExperimentSpecificationFileManager man =
                new ExperimentSpecificationFileManager(eSpec, bundle, eSpecLogger, null, null);

        man.fetchAll();
        assertTrue(man.isReady());

        boolean sawF = false;
        boolean sawDD = false;
        for (UploadSpec upload : eSpec.getUploads()) {
            if (man.isDir(upload.getSource()) && assureEndWithSlash(upload.getSource().getValue()).equals("d/dd/")) {
                Iterator<ExperimentSpecificationFileManager.DirFileBytes> it = man.getDirBytes(upload.getSource());
                assert it.hasNext();
                while(it.hasNext()) {
                    ExperimentSpecificationFileManager.DirFileBytes dirFileBytes = it.next();
                    System.out.println("Iterate dirFileBytes -> "+dirFileBytes.getFullRelativePath()+"  -> base = "+dirFileBytes.getBaseFilename());
                    switch (dirFileBytes.getBaseFilename()) {
                        case "f" : {
                            assertFalse(sawF);
                            assertTrue(sawDD);
                            assertFalse(dirFileBytes.isDir());
                            assertThat(dirFileBytes.getContent(), is(equalTo(new byte[] { 'f', 'f', 'f', '\n' })));
                            assertThat(dirFileBytes.getFullRelativePath(), is(equalTo("f")));
                            sawF = true;
                            break;
                        }
                        case "" : {
                            assertFalse(sawF);
                            assertFalse(sawDD);
                            assertTrue(dirFileBytes.isDir());
                            assertThat(dirFileBytes.getFullRelativePath(), is(equalTo("")));
                            sawDD = true;
                            break;
                        }
                        default: {
                            assertTrue(false, "Bad base filename: "+dirFileBytes.getBaseFilename());
                        }
                    }
                }

            }
        }
        assertTrue(sawDD);
        assertTrue(sawF);
    }





    @Test
    public void testArchiveDirIterator()throws Exception {
        URL archiveUrl = ExperimentSpecificationFileManagerTest.class.getResource("testNestDir.tar.gz");
        Path archivePath = java.nio.file.Paths.get(archiveUrl.toURI());
        //        ESpecBundle bundle = BundleFetcher.fetchArchiveUrl(archiveUrl);
        ESpecBundle bundle = BundleFetcher.fetchArchiveFile(archivePath.toFile());
        //        ESpecBundle bundle = BundleFetcher.fetchArchiveFile(new File("/home/wim/tmp/testNestDir.tar.gz"));

        testArchiveBundleDirs(bundle);
        testArchiveBundleFilenames(bundle);
        testArchiveBundleSubDirs(bundle);

        testDirIterator(bundle);
        testSubDirIterator(bundle);
    }

    @Test
    public void testLocalDirIterator()throws Exception {
        URL archiveUrl = ExperimentSpecificationFileManagerTest.class.getResource("testDir");
        Path archivePath = java.nio.file.Paths.get(archiveUrl.toURI());
        //        ESpecBundle bundle = BundleFetcher.fetchArchiveUrl(archiveUrl);
        ESpecBundle bundle = BundleFetcher.fetchDir(archivePath.toFile());
        //        ESpecBundle bundle = BundleFetcher.fetchArchiveFile(new File("/home/wim/tmp/testNestDir.tar.gz"));

        testArchiveBundleDirs(bundle);
        testArchiveBundleFilenames(bundle);
        testArchiveBundleSubDirs(bundle);

        testDirIterator(bundle);
        testSubDirIterator(bundle);
    }

    @Nonnull
    private static String assureEndWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }
}